import logging
log = logging.getLogger('root')
import numpy as np
#from numba import njit
import traceback
import abc
from xframe.library.pythonLibrary import convert_to_slice_if_possible
from xframe.library.mathLibrary import masked_mean
from xframe.library.mathLibrary import combine_means_2D
from xframe.library.mathLibrary import masked_variance
from xframe.library.mathLibrary import combine_variances_ND
from xframe import Multiprocessing

import warnings


def get_slize_by_bounds(array,bounds):
    start,stop = bounds
    array = np.asarray(array)
    amin = bounds[0]
    amax = bounds[1]
    log.info('array = ')
    if isinstance(amin,(int,float)):
        start = np.sum(array<amin)
    if isinstance(amax,(int,float)):
        stop = np.sum(array<=amax)
    return slice(start,stop,1)
        

def update_dict(default,_dict):
    #log.info('default_keys = {}'.format(default.keys()))
    for key,val in default.items():
        if not (key in _dict):
            _dict[key]=val
    return _dict

class Quantity(abc.ABC):
    _Quantity_default_param = {'data_shape':(16,512,128),'roi_mask':True}
    _Quantity_default_options = {'use_multiprocessing':False}
    def __init__(self,name,parameters = {},options = {}):
        parameters = update_dict(self._Quantity_default_param,parameters)
        options = update_dict(self._Quantity_default_options,options)
        self.name = name
        self.param = parameters
        self.opt = options
        self.data_shape = self.param['data_shape']
        # internal roi_mask_inverted is tue on masked points.
        self.roi_mask_inverted = ~np.asarray(self.param['roi_mask'])
        self.use_multiprocessing = self.opt['use_multiprocessing']
        self.n_processed_chunks=0
        self.data = {}
    @abc.abstractmethod
    def _compute(self,data):
        pass
        
    def compute(self,chunk_id,data):
        log.info('inside name ={} c_id = {} q.n_processed_chunks = {}'.format(self.name,chunk_id,self.n_processed_chunks))
        if chunk_id==self.n_processed_chunks:
            log.info('c_id = {} q.n_processed_chunks = {}'.format(chunk_id,self.n_processed_chunks))
            self._compute(data)
            self.n_processed_chunks+=1
            
    def reset(self):
        self.__init__(self.name,self.param,self.opt)

class DependendQuantity(Quantity):
    quantity_classes = []
    def __init__(self,quantities,name,parameters = {},options = {}):
        super().__init__(name,parameters = parameters,options = options)
        self._check_input_quantities(quantities)
        self.parent_quantities = quantities        
    def _check_input_quantities(self,quantities):
        for q_class,q in zip(self.quantity_classes,quantities.values()):
            if not isinstance(q,q_class):
                raise AssertionError('quantity {} is not of type {}'.format(q,q_class))
        
    def compute(self,chunk_id,data):
        n_processed_chunks = self.n_processed_chunks
        for pq in self.parent_quantities.values():
            pq.compute(chunk_id,data)
            
        if chunk_id==self.n_processed_chunks:
            self._compute(data)
            self.n_processed_chunks+=1
    def reset(self):        
        self.__init__(self.parent_quantities,self.name,self.param,self.opt)

        
class Mean2D(Quantity):
    default_options = {}
    default_param = {}
    def __init__(self,name,options = {}, parameters = {}):
        parameters = update_dict(self.default_param,parameters)
        options = update_dict(self.default_options,options)
        super().__init__(name, options = options,parameters=parameters)
        data_shape = self.data_shape
        log.info('data shape = {}'.format(data_shape))
        
        out_shapes=[data_shape,data_shape]
        out_dtypes = [float,int]
        self._mp_mode = Multiprocessing.MPMode_SharedArray(out_shapes,out_dtypes,reduce_arguments = True)
        
        mean=np.zeros(data_shape,dtype=float)
        mask=np.zeros(data_shape,dtype=bool)
        counts=np.zeros(data_shape,dtype=int)
        #roi_mask is true where data is allowed and self.roi_mask is true where data is masked
        self.data['mean']=mean
        self.data['mask']=mask
        self.data['counts']=counts
    def _compute(self,data_chunk):
        log.info('data shape = {}'.format(self.data['mean'].shape))
        roi_mask = self.roi_mask_inverted
        datasets,masks = data_chunk['data'],data_chunk['mask']
        if self.use_multiprocessing:
            log.error('Multiprocessing not implemented yet. Switch to normal mode.')
            #mean,mask,counts = self.calc_mean_mp(datasets,masks)
            mean,mask,counts = self.calc_mean_mp(datasets,masks)
        else:
            mean,mask,counts = self.calc_mean_mp(datasets,masks)
        #mean[roi_mask]=0
        #counts[roi_mask]=0
        log.info('mask shape ={} roi_mask shape = {}'.format(mask.shape,roi_mask.shape))
        #mask[roi_mask]=False
        self.data['mean'] = mean
        self.data['counts'] = counts
        self.data['mask'] = mask
    def calc_mean_mp(self,datasets,masks):
        old_mean,old_counts,old_mask = self.data['mean'],self.data['counts'],self.data['mask']
        def _mean(frame_ids,datasets,masks,**kwargs):
            out_mean,out_counts = kwargs['outputs']
            out_ids = kwargs['output_ids']
            tmp_masks = masks[frame_ids].copy()
            tmp_masks[:,self.roi_mask_inverted]=False
            mean,counts  = masked_mean(datasets[frame_ids],tmp_masks,axis=0)
            out_mean[out_ids] = mean
            out_counts[out_ids] = counts
        n_frames = len(datasets)
        data_means,data_counts = Multiprocessing.comm_module.request_mp_evaluation(_mean,self._mp_mode,input_arrays=[np.arange(n_frames)],const_inputs=[datasets,masks],call_with_multiple_arguments=True,split_mode='modulus')
        log.info(f'mean {data_means.dtype} counts = {data_counts.dtype}')
        data_mean,data_count = combine_means_2D(data_means,data_counts)
        log.info(f'mean {data_mean.dtype} counts = {data_count.dtype}')
        mean,counts = combine_means_2D((data_mean,old_mean),(data_count,old_counts))
        log.info(f'mean {mean.dtype} counts = {counts.dtype}')
        mask = (counts != 0)
        return mean,mask,counts
   
class Std2DSimple(DependendQuantity):
    quantity_classes = [Mean2D]
    def __init__(self,quantities,name,parameters = {},options = {}):
        super().__init__(quantities,name,parameters = parameters,options = options)
        data_shape = self.data_shape
        log.info('data shape = {}'.format(data_shape))
        
        out_shapes=[data_shape,data_shape]
        out_dtypes = [float,int]
        self._mp_mode = Multiprocessing.MPMode_SharedArray(out_shapes,out_dtypes,reduce_arguments = True)
        
        mean=np.zeros(data_shape,dtype=float)
        mask=np.zeros(data_shape,dtype=bool)
        counts=np.zeros(data_shape,dtype=int)
        #roi_mask is true where data is allowed and self.roi_mask is true where data is masked
        self.data['mean_square']=mean
        self.data['mask']=mask
        self.data['counts']=counts

        self.mean_obj = tuple(self.parent_quantities.values())[0]
        
    def _compute(self,data_chunk):
        mean_square,counts = self.calc_mean_square_mp(data_chunk['data'],data_chunk['mask'])
        self.data['mean_square']=mean_square
        self.data['mask']=self.mean_obj.data['mask'].copy()
        self.data['counts']=counts
        var = mean_square-self.mean_obj.data['mean']**2
        var[var<0]=0
        self.data['std']=np.sqrt(var)
        
    def calc_mean_square_mp(self,datasets,masks):
        old_mean,old_counts = self.data['mean_square'],self.data['counts']
        def _mean(frame_ids,datasets,masks,**kwargs):
            out_mean,out_counts = kwargs['outputs']
            out_ids = kwargs['output_ids']
            tmp_masks = masks[frame_ids].copy()
            tmp_masks[:,self.roi_mask_inverted]=False
            out_mean[out_ids],out_counts[out_ids] = masked_mean(datasets[frame_ids]**2,tmp_masks,axis=0)
        n_frames = len(datasets)
        data_means,data_counts = Multiprocessing.comm_module.request_mp_evaluation(_mean,self._mp_mode,input_arrays=[np.arange(n_frames)],const_inputs=[datasets,masks],call_with_multiple_arguments=True,split_mode='modulus')
        data_mean,data_count = combine_means_2D(data_means,data_counts)
        mean,counts = combine_means_2D((data_mean,old_mean),(data_count,old_counts))
        return mean,counts
    
class Std2D(Quantity):
    default_options = {}
    default_param = {}
    def __init__(self,name,options = {}, parameters = {}):
        parameters = update_dict(self.default_param,parameters)
        options = update_dict(self.default_options,options)
        super().__init__(name, options = options,parameters=parameters)
        data_shape = self.data_shape
        log.info('data shape = {}'.format(data_shape))
        std=np.zeros(data_shape,dtype=float)
        mask=np.zeros(data_shape,dtype=bool)
        counts=np.zeros(data_shape,dtype=int)
        #roi_mask is true where data is allowed and self.roi_mask is true where data is masked
        self.data['std']=std
        self.data['var']=std.copy()
        self.data['mean']=std.copy()
        self.data['mask']=mask
        self.data['counts']=counts
        
    def _compute(self,data_chunk):
        log.info('data shape = {}'.format(self.data['mean'].shape))
        roi_mask = self.roi_mask_inverted
        datasets,masks = data_chunk['data'],data_chunk['mask']
        if self.use_multiprocessing:
            log.error('Multiprocessing not implemented yet. Switch to normal mode.')
            #mean,mask,counts = self.calc_mean_mp(datasets,masks)
            var,mean,mask,counts = self.calc_mean(datasets,masks)
        else:
            var,mean,mask,counts = self.calc_mean(datasets,masks)
        #mean[roi_mask]=0
        #counts[roi_mask]=0
        log.info('mask shape ={} roi_mask shape = {}'.format(mask.shape,roi_mask.shape))
        mask[roi_mask]=False
        self.data['var']=var
        self.data['std']=np.sqrt(var)
        self.data['mean'] = mean
        self.data['counts'] = counts
        self.data['mask'] = mask
    def calc_mean_mp(self,datasets,masks):
        pass
    def calc_mean(self,datasets,masks,**kwargs):
        old_var,old_mean,old_counts,old_mask = self.data['var'],self.data['mean'],self.data['counts'],self.data['mask']
        #log.info('datasets shape = {} masks shape = {}'.format(datasets.shape,masks.shape))
        new_var,new_counts = masked_variance(datasets,masks,axis=0)
        new_mean,_ = masked_mean(datasets,masks,axis=0)
        var,mean,counts = combine_variances_ND((new_var,old_var),(new_mean,old_mean),(new_counts,old_counts))
        mask = np.sum(masks,axis=0).astype(bool) | old_mask
        return [var,mean,mask,counts]
    
class Maximum2D(Quantity):
    default_param = {}
    default_options = {}
    def __init__(self,name,parameters = {}, options = {}):
        parameters = update_dict(self.default_param,parameters)
        options = update_dict(self.default_options,options)
        super().__init__(name,options = options,parameters = parameters)
        data_shape = self.data_shape

        out_shapes=[data_shape,data_shape]
        out_dtypes = [float,bool]
        self._mp_mode = Multiprocessing.MPMode_SharedArray(out_shapes,out_dtypes,reduce_arguments = True)
        
        maximum = np.zeros(data_shape,dtype=float)
        mask = np.zeros(data_shape,dtype=bool)
        self.data['maximum']=maximum
        self.data['mask']=mask
    
    def _compute(self,data_chunk):
        roi_mask = self.roi_mask_inverted
        maximum,mask = self.calc_max_mp(data_chunk['data'],data_chunk['mask'])
        log.info('maximum shape ={} mask shape = {}'.format(maximum.shape,mask.shape))
        self.data['maximum']=maximum
        self.data['mask']=mask
    
    def calc_max_mp(self,data,masks):
        log.info('call multiprocess')
        n_frames = len(data)
        max_per_process,mask_per_process = Multiprocessing.comm_module.request_mp_evaluation(self._max_worker,self._mp_mode,input_arrays=[np.arange(n_frames)],const_inputs=[data,masks,self.roi_mask_inverted], call_with_multiple_arguments = True,split_mode = 'modulus')
        log.info('per process shapes = {} | {}'.format(max_per_process.shape,mask_per_process.shape))
        #if len(data_per_process[0])==2:
        #    max_per_process = np.array([d[0] for d in data_per_process.values()])
        #    mask_per_process = np.array([d[1] for d in data_per_process.values()])
        #else:
        #    max_per_process = np.array([d[0][0] for d in data_per_process.values()])
        #    mask_per_process = np.array([d[0][1] for d in data_per_process.values()])
        #data_per_process = np.asarray(data_per_process)
        #log.info('max data shape = {} '.format(mask_per_process.shape))
        data_max = np.max(max_per_process,axis = 0,where=mask_per_process,initial=0)
        data_mask = np.sum(mask_per_process,axis=0).astype(bool)
        
        new_maximum = np.maximum(data_max,self.data['maximum'])
        new_mask = data_mask | self.data['mask']
        return new_maximum,new_mask

    @staticmethod
    def _max_worker(frame_ids,data,masks,roi_mask_inverted,**kwargs):
        out_max,out_masks = kwargs['outputs']
        out_ids = kwargs['output_ids']
        tmp_masks = np.array(masks[frame_ids])
        tmp_masks[:,roi_mask_inverted]=False
        #log.info('shapes = {} | {} | out_max {} | out ids {}'.format(data.shape,masks.shape,out_max.shape,out_ids))
        out_max[out_ids] = np.max(data[frame_ids],axis = 0,where=tmp_masks,initial=0)
        out_masks[out_ids] = np.sum(tmp_masks,axis = 0).astype(bool)
    
class radial_profiles(Quantity):
    def __init__(self,quantities,name,parameters = {},options = {}):
        super().__init__(name,options = options,parameters = parameters)
        data_shape = self.data_shape

        out_shapes=[data_shape,data_shape]
        out_dtypes = [float,bool]
        self._mp_mode = Multiprocessing.MPMode_SharedArray(out_shapes,out_dtypes,reduce_arguments = True)

        self.regridder = self.param['regridder']
        
        self.qs = new_grid[:,0]
        self.phis = new_grid[0,:]
        quantity_shape = self.qs.shape
        
        super().__init__(quantities,name,parameters = parameters,options = options)

        self.roi_modules=roi_modules
        self.data_modules = data_modules
        #roi_mask is true where data is allowed and self.roi_mask is true where data is masked        
        self.roi_mask =  ~np.asarray(roi_mask)
        
        saxs = np.zeros(data_shape,dtype = float)
        mask = np.zeros(data_shape,dtype = bool)
        self.data = {'saxs':saxs,'mask':mask}

        self.use_multiprocessing=use_multiprocessing
        self.background_scale = False
class Saxs(DependendQuantity):
    quantity_classes = [Mean2D]

    def __init__(self,quantities,name,parameters = {},options = {}):
    #def __init__(self,mean_2d,regridder,data_shape = (16,512,128),roi_mask = False,roi_modules = True,data_modules= True,use_multiprocessing=False,options = {}):
        self.regridder = regridder
        self.qs = new_grid[:,0]
        self.phis = new_grid[0,:]
        quantity_shape = self.qs.shape
        
        super().__init__(quantities,name,parameters = parameters,options = options)

        self.roi_modules=roi_modules
        self.data_modules = data_modules
        #roi_mask is true where data is allowed and self.roi_mask is true where data is masked        
        self.roi_mask =  ~np.asarray(roi_mask)
        
        saxs = np.zeros(data_shape,dtype = float)
        mask = np.zeros(data_shape,dtype = bool)
        self.data = {'saxs':saxs,'mask':mask}

        self.use_multiprocessing=use_multiprocessing
        self.background_scale = False
    def _compute(self,data_chunk):
        datasets,masks = data_chunk['data'],data_chunk['mask']
        if self.use_multiprocessing:
            log.error('Multiprocessing not implemented yet. Switch to normal mode.')
            #mean,mask,counts = self.calc_saxs_mp(datasets,masks)
            saxs,mask,counts = self.calc_saxs()
        else:
            saxs,mask,counts = self.calc_saxs()
        self.data['mean'] = mean
        self.data['counts'] = counts
        self.data['mask'] = mask

    def calc_saxs_mp(self):
        pass
        
    def calc_saxs(self):
        opt = self.opt
        regridder = self.regridder
        mean2d = self.parent_quantities[0]
        roi_modules = self.roi_modules
        data_modules = self.data_modules
        roi_mask = self.roi_mask

        
        mean_data=np.zeros(data_shape,dtype = float)
        mask_data=np.zeros(data_shape,dtype = bool)
        tmp_mean = mean_data[roi_modules]
        tmp_mask = mask_data[roi_modules]
        tmp_mean[roi_mask] = mean2d.data['mean'][data_modules][roi_mask].flatten()
        tmp_mask[roi_mask] = mean2d.data['mean'][data_modules][roi_mask].flatten()
        mean_data[roi_modules] = tmp_mean
        mask_data[roi_modules] = tmp_mask
        
        mean_on_polar_grid,mask_on_polar_grid = regridder.regrid(mean_data,mask_data.astype(bool),np.arange(16))
        
        saxs,counts = masked_mean(mean_on_polar_grid,mask_on_polar_grid,axis = 1)
        mask = np.sum(mask_on_polar_grid,axis = 1).astype(bool)

        if opt.get('background',{'subtract':False})['subtract']:
            saxs = self.subtract_background(saxs)
        return saxs,counts,mask


    def subtract_background(self,saxs):
        opt = self.opt
        bg_saxs = opt['background']['saxs']
        bg_saxs_mask = opt['background']['mask']
        scaled_bg_saxs = self.scale_background_to_data(saxs,bg_saxs)
        return saxs-scaled_bg_saxs
    def scale_background_to_data(self,data_saxs,bg_saxs):
        scale_opt = self.opt['background']['scale_opt']
        diff_metric = scale_opt['diff_metric']
        q_range = scale_opt['q_range']
        
        #calculate saxs slice corresponding to scaling q_range
        qs = self.qs
        #log.info('use q_range for scaling = {}'.format(q_range))
        q_slice = get_slize_by_bounds(qs,q_range)
        
        #determine bounds for the scaling factor in the 1d minimization.
        #That is: 1 sigma interval around mean value of data_saxs/background_saxs
        non_zero_mask = (bg_saxs!=0)
        #log.info('data type = {} backgorund type = {}'.format(data_saxs.shape,background_saxs.shape))
        scales_per_q=data_saxs[non_zero_mask]/background_saxs[non_zero_mask]
        diff_mean = np.mean(scales_per_q)
        diff_std = np.std(scales_per_q)
        #prefiousely std of background - data
        bounds = (diff_mean-diff_std,diff_mean+diff_std)
        
        def diff(scale):
            return diff_metric(np.abs(data_saxs[q_slice] - scale*background_saxs[q_slice]))
        scale = minimize_scalar(diff,method='bounded',bounds = bounds,options={'maxiter':scale_opt['max_iterations'],'xatol':scale_opt['abs_tolerance']}).x
        self.background_scale = scale
        return bg_saxs*scale


class Sum1D(Quantity):
    def __init__(self,name,parameters = {},options = {}):
        super().__init__(name,options = options,parameters = parameters)
        self.x_axis_type = self.opt['x_axis']
        self.data['pixels_per_frame']=np.atleast_1d(np.prod(self.data_shape)-np.sum(self.roi_mask_inverted))
        self.data['values'] = np.array([],dtype = float)
        self.data['points'] = np.array([],dtype = float)
        self.data['frame_ids'] = np.array([],dtype = float)
        self.data['mask'] = np.array([],dtype = bool)
        self.data['point_type']=self.x_axis_type
        

    def _compute(self,data_chunk):
        datasets,masks,points = data_chunk['data'],data_chunk['mask'],data_chunk[self.x_axis_type]
        self.data['frame_ids'] = np.concatenate((self.data['frame_ids'],data_chunk['frame_ids']))
        #roi_mask = self.roi_mask_inverted
        if self.use_multiprocessing:
            log.error('Multiprocessing not implemented yet. Switch to normal mode.')
            #mean,mask,counts = self.calc_mean_mp(datasets,masks)
            values,masks = self.reduce_1d_mp(datasets,masks,points)
        else:
            values,masks = self.reduce_1d_mp(datasets,masks,points)
        self.data['values'] = np.concatenate((self.data['values'],values))
        self.data['points'] = np.concatenate((self.data['points'],points))
        self.data['mask'] = np.concatenate((self.data['mask'],masks))
        
    def reduce_1d_mp(self,data,masks,new_points):
        roi_mask = self.param['roi_mask']
        def _sum(frame_ids,data,masks,**kwargs):
            n_frames = len(frame_ids)
            out_sums,out_masks = kwargs['outputs']
            out_ids = kwargs['output_ids']
            flat_data = data[frame_ids].reshape(n_frames,-1)
            flat_masks = (masks[frame_ids]*roi_mask[None,:]).reshape(n_frames,-1)
            
            outs = np.sum(flat_data,where=flat_masks,axis = 1)
            outm = np.sum(flat_masks,axis = 1).astype(bool)
            out_sums[out_ids] = outs
            out_masks[out_ids] = outm
        n_patterns = len(data)
        mp_mode = Multiprocessing.MPMode_SharedArray([n_patterns,n_patterns],[float,bool])
        sums,masks = Multiprocessing.comm_module.request_mp_evaluation(_sum,mp_mode,input_arrays=[np.arange(n_patterns)],const_inputs = [data,masks], call_with_multiple_arguments = True)
        return sums,masks



import numpy as np
import os
os.chdir('/gpfs/exfel/theory_group/user/berberic/Programs/xframe')
from xframe.database.euxfel import default_DB
from xframe.library.gridLibrary import GridFactory
from xframe.presenters.matplolibPresenter import plot1D,heat2D_multi,heat2D,heatPolar2D,agipd_heat
from xframe.library.mathLibrary import circularHarmonicTransform_real_forward,cartesian_to_spherical,spherical_to_cartesian
from xframe.library.mathLibrary import circularHarmonicTransform_complex_forward

import logging
#from xframe.plugins.fxs3046_online.analysisLibrary.regrid2 import SimpleRegridder2D,AgipdRegridderSimple

from scipy.spatial.kdtree import KDTree
from scipy.ndimage import gaussian_filter

log=logging.getLogger('root')

init = True
if init:
    from xframe.startup_routines import dependency_injection
    dependency_injection()
    from xframe.startup_routines import load_recipes
    from xframe.control.Control import Controller
    analysis_sketch,exp_sketch = load_recipes(ana_name='fxs3046_online.generate_plots_2', ana_settings='saxs_65_-bg',exp_name='fxs3046_online',exp_settings = 'default_proc')
    controller = Controller(analysis_sketch,experiment_sketch = exp_sketch)
    controller.chooseJobSingleProcess()
    from xframe import database
    from xframe import Multiprocessing

from xframe import database
db = database.analysis


    
comm = Multiprocessing.comm_module
g = comm.get_geometry()

base_path = '/gpfs/exfel/theory_group/user/berberic/fxs3046/test/regrid2/'

f_data = db.load('/gpfs/exfel/theory_group/user/berberic/p3046/data/fig/run_64/2022-09-02_23:12:32/bg_data_proc.h5')


mean_dict = f_data['mean_2d']
bg_data={}
bg_data['mean'] = mean_dict['mean']
log.info('mean total 2 ={}'.format(np.mean(bg_data['mean'])))
bg_data['mask'] = mean_dict['mask']
#bg_data['settings_dir'] = self.opt['out_settings_path']
bg_data['run'] = 64
bg_data['n_frames']  =  275379
db.save('background',bg_data)



from xframe.settings import general
import logging
log = logging.getLogger('root')

try:
    import psutil
    class PsutilPlugin:
        @staticmethod
        def get_free_memory():
            return psutil.virtual_memory().available
except Exception as e:
    log.warning('psutils not installed. Use RAM value specified in settings.general instead.')
    class PsutilPlugin:
        @staticmethod
        def get_free_memory():
            return general.RAM



import importlib
import traceback
import logging
log = logging.getLogger('root')

modules_to_import  = ['matplotlibPresenter','openCVPresenter']
for module in modules_to_import:
    try:
        if isinstance(module,(list,tuple)):
            module_name = 'xframe.presenters.'+module[0]
            globals()[module[1]] = importlib.import_module(module_name)
        else:
            splitted_name = module.split('.')
            module_name = 'xframe.presenters.'+module
            if len(splitted_name) > 1:                
                globals()[splitted_name[-1]] = importlib.import_module(module_name)
            else:
                globals()[module] = importlib.import_module(module_name)
                
    except Exception as e:
        traceback.print_exc()
        log.error('Caught exeption while loading {} with content: {}'.format(module_name,e))



from ruamel.yaml import YAML
from ruamel.yaml.comments import CommentedMap,CommentedSeq
from xframe.library.pythonLibrary import DictNamespace
from xframe.settings.tools import SettingsConverter
import numpy as np
import logging
import traceback

log=logging.getLogger('root')
yaml = YAML()

class YAML_access:
    @staticmethod
    def save(path, data:dict,convert=False,**kwargs):
        with open(path,'w') as _file:
            if convert:
                yaml.dump(SettingsConverter.convert(data),_file)
            else:
                yaml.dump(data,_file)
    @staticmethod
    def load(path,**kwargs):
        with open(path,'r') as _file:
            data = yaml.load(_file.read())
        if isinstance(data,type(None)):
            data = {}
        return data
    @staticmethod
    def format_settings_dict(settings):
        return SettingsConverter.convert(settings)




from . import fxs_invariant_tools
from . import hankel_transforms
from . import harmonic_transforms
from . import ft_grid_pairs
from . import fxs_Projections as fxs_projections
from . import cross_correlation
from . import fxs_IO_methods
from . import misk




import numpy as np
from xframe.library import mathLibrary as mLib
from xframe import log
import scipy as sp
log=log.setup_custom_logger('root','INFO')

#A = np.random.rand(3*20*10).reshape(3,20,10)
size = 100
n_systems = 3
A = np.diag(np.exp(-np.arange(0,size)*np.log(1e+5)/size))
#A[0,-1]=100
x_exact = np.arange(0.9,1,0.1/size)[::-1]
b_free = A.dot(x_exact)
b = b_free[None,:] + np.random.normal(scale = 0.1,size = n_systems*size).reshape(n_systems,size)

A = np.stack((A,)*n_systems)


'''
    Finds x that minimizes $|| A x -B ||^2 - ||\lambda Id x ||^2 $
'''
if A.ndim < 3:
    A = A[None,...]
    b = b[None,...]
try:
    #lambd = mLib.approximate_tikhonov_parameters(AA,bb)
    lambd = mLib.approximate_tikhonov_parameters(A,b)
    #lambd = np.array([0.0])
except Exception:
    print('fuu')
    lambd = np.array([0.0])

allow_offset = True
#from sklearn import linear_model
#reg2 = linear_model.RidgeCV(alphas=np.logspace(-6, 6, 13),gcv_mode='svd',fit_intercept=False,normalize = False)
#reg = linear_model.Ridge(alpha=lambd[0],solver='svd',fit_intercept=allow_offset,normalize = False)
#reg.fit(A[0],b[0])
#reg2.fit(A[0],b[0])
#x1 = np.copy(reg.coef_)
#x2 = reg2.coef_
#new_b = A[0].dot(x1) + reg.intercept_
#
#lambd2 = mLib.approximate_tikhonov_parameters(A[0],new_b)
#reg3 = linear_model.Ridge(alpha=lambd2[0],solver='svd',fit_intercept=allow_offset,normalize = False)
#x3 = reg3.fit(A[0],new_b).coef_

#x,o= mLib.tikhonov_regularization(A,b,lambd,allow_offset=allow_offset)
xx,oo= mLib.optimal_tikhonov_regularization(A,b,allow_offset=allow_offset,iterations = 1)
xxx,ooo= mLib.optimal_tikhonov_regularization(A,b,allow_offset=allow_offset,iterations = 1)

new_b2 = A[0].dot(xx[0])+oo[0]
new_bs = np.sum(A*xx[:,None,:],axis = -1)+oo[:,None]



import cv2 as cv
import numpy as np
from xframe.database.interfaces import OpenCVInterface
from xframe.presenters.openCVPresenter import OpenCVInterface as PresenterInterface

class CV_Plugin(OpenCVInterface,PresenterInterface):
    colormaps={
        'autumn':cv.COLORMAP_AUTUMN,
        'bone':cv.COLORMAP_BONE,
        'jet':cv.COLORMAP_JET,
        'winter':cv.COLORMAP_WINTER,
        'rainbow':cv.COLORMAP_RAINBOW,
        'ocean': cv.COLORMAP_OCEAN,
        'summer': cv.COLORMAP_SUMMER,
        'spring': cv.COLORMAP_SPRING,
        'cool': cv.COLORMAP_COOL,
        'hsv': cv.COLORMAP_HSV,
        'pink': cv.COLORMAP_PINK,
        'hot': cv.COLORMAP_HOT,
        'parula': cv.COLORMAP_PARULA,
        'magma': cv.COLORMAP_MAGMA,
        'inferno': cv.COLORMAP_INFERNO,
        'plasma': cv.COLORMAP_PLASMA,
        'viridis': cv.COLORMAP_VIRIDIS,
        'cvidis': cv.COLORMAP_CIVIDIS,
        'twilight': cv.COLORMAP_TWILIGHT,
        'twilight_shfted': cv.COLORMAP_TWILIGHT_SHIFTED,
        'turbo': cv.COLORMAP_TURBO,
        'deep_green':cv.COLORMAP_DEEPGREEN
    }

    @classmethod
    def get_polar_image_data(cls,data,n_pixels=False):
        Nr,Nphi=data.shape
        pdata=data.copy()
        _min = pdata.min()
        if isinstance(n_pixels,bool):
            n_pixels=Nr*2
        radius = int(n_pixels//2)
        center = (int(n_pixels//2),int(n_pixels//2))
            
        pic = cv.warpPolar(pdata.T,center=center,maxRadius=radius,dsize=(n_pixels,n_pixels),flags=cv.WARP_INVERSE_MAP + cv.WARP_POLAR_LINEAR)
        return pic
    

    @classmethod
    def get_polar_image(cls,data,n_pixels=False,scale='lin',colormap='viridis',vmin=False,vmax=False,print_colorscale=False,transparent_backgound=False):
        Nr,Nphi=data.shape
        pdata=data.copy()
        use_log_scale= (scale=='log')
        if use_log_scale:
            pdata[pdata<=0]=1e-15
            pdata=np.log10(pdata)         
        

        max_d = pdata.max()
        min_d = pdata.min()
        if isinstance(vmin,bool):
            vmin=min_d
        else:
            if use_log_scale:
                vmin=np.log10(np.abs(vmin))
            pdata[pdata<vmin]=vmin
            
        if isinstance(vmax,bool):
            vmax=max_d
        else:
            if use_log_scale:
                vmax=np.log10(np.abs(vmax))
            pdata[pdata>vmax]=vmax
        zero_in_range= (0>=vmin) and (0<=vmax)
        if zero_in_range:
            bg_value = 0
        else:
            bg_value = vmin

        enlarged_Nr = int(Nr*np.sqrt(2))
        temp_pic = np.zeros((enlarged_Nr,Nphi),dtype=float)
        bg_pic = np.zeros((enlarged_Nr,Nphi),dtype=float)
        temp_pic[:Nr]=pdata
        temp_pic[Nr:]=bg_value
        bg_pic[Nr:]=1
        
        if isinstance(n_pixels,bool):
            n_pixels=Nr*2
        radius = int(n_pixels/np.sqrt(2))
        center = (int(n_pixels//2),int(n_pixels//2))
            
        pic = cv.warpPolar(temp_pic.T,center=center,maxRadius=radius,dsize=(n_pixels,n_pixels),flags=cv.WARP_INVERSE_MAP + cv.WARP_POLAR_LINEAR)
        bg_mask = cv.warpPolar(bg_pic.T,center=center,maxRadius=radius,dsize=(n_pixels,n_pixels),flags=cv.WARP_INVERSE_MAP + cv.WARP_POLAR_LINEAR)==1
        
        pic-=vmin
        pic*=255/(vmax-vmin)
        pic=pic.astype(np.uint8)
        pic= cv.applyColorMap(pic, cls.colormaps.get(colormap,cv.COLORMAP_VIRIDIS))
        if print_colorscale:
            h_black = int(pic.shape[0]*0.05)
            h_color_bar = int(pic.shape[0]*0.05)
            cscale=np.zeros((h_black+h_color_bar,pic.shape[0]),dtype=np.uint8)
            cscale[h_black:]=np.arange(pic.shape[0])*255/pic.shape[0]
            cscale = cv.applyColorMap(cscale, cls.colormaps.get(colormap,cv.COLORMAP_VIRIDIS))
            cscale[:h_black]=0

            if use_log_scale:
                vmin_txt = f"{10**(vmin):1.2e}"
                vmax_txt = f"{10**(vmax):1.2e}"
            else:
                vmin_txt = f"{vmin:1.2e}"
                vmax_txt = f"{vmax:1.2e}"
            WHITE = (255,255,255)
            font = cv.FONT_HERSHEY_SIMPLEX
            scale = 0.03 # this value can be from 0 to 1 (0,1] to change the size of the text relative to the image
            font_size = n_pixels/(25/scale)
            font_color = WHITE
            font_thickness = int(font_size*2)
            img_text = cv.putText(cscale, vmin_txt, (int(n_pixels*0),h_color_bar-int(h_black*0.2)), font, font_size, font_color, font_thickness, cv.LINE_AA)
            img_text = cv.putText(cscale, vmax_txt, (int(n_pixels*0.80),h_color_bar-int(h_black*0.2)), font, font_size, font_color, font_thickness, cv.LINE_AA)
            pic=np.concatenate((pic,cscale),axis=0)
            

        if transparent_backgound:
            # First create the image with alpha channel
            pic_rgba = cv.cvtColor(pic, cv.COLOR_RGB2RGBA)

            # Then assign the mask to the last channel of the image
            pic_rgba[:n_pixels, :n_pixels, 3] = ((~bg_mask).astype(np.uint8)*255)
            pic=pic_rgba.copy()

        return pic


    @classmethod
    def get_cart_image(cls,data,scale='lin',colormap='viridis',vmin=False,vmax=False,print_colorscale=False):
        Nx,Ny=data.shape
        pdata=data.copy()
        n_pixels = Nx
        use_log_scale= (scale=='log')
        if use_log_scale:
            pdata[pdata<=0]=1e-15
            pdata=np.log10(pdata)         
        

        max_d = pdata.max()
        min_d = pdata.min()
        if isinstance(vmin,bool):
            vmin=min_d
        else:
            if use_log_scale:
                vmin=np.log10(np.abs(vmin))
            pdata[pdata<vmin]=vmin
            
        if isinstance(vmax,bool):
            vmax=max_d
        else:
            if use_log_scale:
                vmax=np.log10(np.abs(vmax))
            pdata[pdata>vmax]=vmax
        zero_in_range= (0>=vmin) and (0<=vmax)
        if zero_in_range:
            bg_value = 0
        else:
            bg_value = vmin

        pic = pdata
        
        pic-=vmin
        pic*=255/(vmax-vmin)
        pic=pic.astype(np.uint8)
        pic= cv.applyColorMap(pic, cls.colormaps.get(colormap,cv.COLORMAP_VIRIDIS))
        if print_colorscale:
            h_black = int(pic.shape[0]*0.05)
            h_color_bar = int(pic.shape[0]*0.05)
            cscale=np.zeros((h_black+h_color_bar,pic.shape[0]),dtype=np.uint8)
            cscale[h_black:]=np.arange(pic.shape[0])*255/pic.shape[0]
            cscale = cv.applyColorMap(cscale, cls.colormaps.get(colormap,cv.COLORMAP_VIRIDIS))
            cscale[:h_black]=0

            if use_log_scale:
                vmin_txt = f"{10**(vmin):1.2e}"
                vmax_txt = f"{10**(vmax):1.2e}"
            else:
                vmin_txt = f"{vmin:1.2e}"
                vmax_txt = f"{vmax:1.2e}"
            WHITE = (255,255,255)
            font = cv.FONT_HERSHEY_SIMPLEX
            scale = 0.03 # this value can be from 0 to 1 (0,1] to change the size of the text relative to the image
            font_size = n_pixels/(25/scale)
            font_color = WHITE
            font_thickness = int(font_size*2)
            img_text = cv.putText(cscale, vmin_txt, (int(n_pixels*0),h_color_bar-int(h_black*0.2)), font, font_size, font_color, font_thickness, cv.LINE_AA)
            img_text = cv.putText(cscale, vmax_txt, (int(n_pixels*0.80),h_color_bar-int(h_black*0.2)), font, font_size, font_color, font_thickness, cv.LINE_AA)
            pic=np.concatenate((pic,cscale),axis=0)            
        return pic

    @classmethod
    def save(cls,path,image):
        path=path[:-3]+'.png'
        cv.imwrite(path, image)
        cv.waitKey(0)
    def load(cls,path):
        image = cv.imread(path)
        cv.waitKey(0)
        return image



import numpy as np
import logging
from xframe.library.mathLibrary import get_spherical_harmonic_transform_obj
from xframe.library.mathLibrary import circularHarmonicTransform_complex_forward
from xframe.library.mathLibrary import circularHarmonicTransform_complex_inverse
from xframe.library.mathLibrary import circularHarmonicTransform_real_forward
from xframe.library.mathLibrary import circularHarmonicTransform_real_inverse
log=logging.getLogger('root')

class HarmonicTransform:    
    def __init__(self,data_type,opt):
        self.data_type=data_type
        self.opt=opt
        self.dim=opt['dimensions']
        ht,iht,grid_param,trf_by_indices = self.chose_transforms()
        self.transforms_by_indices = trf_by_indices
        self.forward = ht
        self.inverse = iht        
        self.grid_param = grid_param

    def chose_transforms(self):
        opt=self.opt
        data_type=self.data_type
        dim=self.dim
        
        if dim == 2:
            #size=opt['n_angular_points']
            size = opt['n_phi']
            grid_opt={}        
            if data_type=='complex':
                ht_forward = circularHarmonicTransform_complex_forward
                ht_inverse = circularHarmonicTransform_complex_inverse
                
            elif data_type == 'real':
                ht_forward = circularHarmonicTransform_real_forward            
                def ht_inverse(data):
                    return circularHarmonicTransform_real_inverse(data,size)                
            trf_by_indices={'m':{'forward':ht_forward,'inverse':ht_inverse}}
            
        elif dim == 3:
            l_max=int(opt['max_order'])
            anti_aliazing_degree = opt.get('anti_aliazing_degree',False)
            n_phi = opt['n_phi']
            n_theta = opt['n_theta']
            indices=opt.get('indices','lm')
            if data_type == 'complex':
                if isinstance(anti_aliazing_degree,bool):
                    sh=get_spherical_harmonic_transform_obj(l_max,mode='complex',n_phi=n_phi,n_theta=n_theta)
                else:
                    sh=get_spherical_harmonic_transform_obj(l_max,mode='complex',anti_aliazing_degree = anti_aliazing_degree ,n_phi= n_phi,n_theta=n_theta)
                self.m_indices = sh.cplx_m_indices
                self.l_indices = sh.cplx_l_indices
                self.m = sh.m
                self.l = sh.l
                self.m_split_indices = sh.cplx_m_split_indices
                self.l_split_indices = sh.cplx_l_split_indices
                self.n_coeff = sh.n_coeff
            elif data_type == 'real':
                if isinstance(anti_aliazing_degree,bool):
                    sh=get_spherical_harmonic_transform_obj(l_max,mode='real')
                else:
                    sh=get_spherical_harmonic_transform_obj(l_max,mode='real',anti_aliazing_degree = anti_aliazing_degree)
                raise NotImplementedError('real spherical harmonic transforms are not implemented right now.')
            self.test = sh.test
            

            trf_by_indices={'lm':{'forward':sh.forward_l,'inverse':sh.inverse_l},'ml':{'forward':sh.forward_m,'inverse':sh.inverse_m},'direct':{'forward':sh.forward_d,'inverse':sh.inverse_d}}
            
            ht_forward=trf_by_indices[indices]['forward']
            ht_inverse=trf_by_indices[indices]['inverse']
            grid_opt={'phis':sh.phi,'thetas':sh.theta}
        return ht_forward,ht_inverse,grid_opt,trf_by_indices








import pytest
import shutil
import os
import sys
import subprocess
import numpy as np
import importlib
from tlib import copy,save,load,create_path_if_nonexistent
import xframe
config_path = os.path.expanduser('~/.xframe/config.py')
config_backup_path = os.path.expanduser('~/.xframe/backup_config.py')

tmp_config_path = ''
_path = ''

@pytest.fixture(scope='class')
def tmp_path(tmpdir_factory):
    path = tmpdir_factory.mktemp('xtest'+str(np.random.rand()))
    return path

@pytest.fixture(scope='class',autouse=True)
def set_temp_home(tmp_path):    
    config_file_exists = os.path.exists(config_path)
    if config_file_exists:
        initial_config = load(config_path)
        print(f'initial config content = {initial_config}')
        copy(config_path,config_backup_path)
    global tmp_config_path
    tmp_config_path = f'{tmp_path}/.xframe/'
    global _path
    _path=tmp_path
    print(f'given path exists: {os.path.exists(tmp_path)}')
    yield
    if config_file_exists:
        copy(config_backup_path,config_path)
    print(f'config after test = {load(config_path)}')
    
def home_files_exist():
    paths=[
        'projects/fxs',
        'projects/tutorial',
        'experiments/SPB',
        'settings/projects/fxs/correlate/tutorial.yaml',
        'settings/projects/fxs/correlate/default_0.01.yaml',
        'settings/projects/fxs/simulate_ccd/tutorial.yaml',
        'settings/projects/fxs/simulate_ccd/default_0.01.yaml',
        'settings/projects/fxs/extract/tutorial.yaml',
        'settings/projects/fxs/extract/default_0.01.yaml',
        'settings/projects/fxs/reconstruct/tutorial.yaml',
        'settings/projects/fxs/reconstruct/default_0.01.yaml',
        'settings/projects/fxs/average/tutorial.yaml',
        'settings/projects/fxs/average/default_0.01.yaml',
        'settings/experiments/SPB/tutorial.yaml',
        'settings/experiments/SPB/default_0.01.yaml',
    ]
    global tmp_config_path
    for path in paths:
        _path= os.path.join(tmp_config_path,path)
        exists = os.path.exists(_path)
        if not exists:
            assert _path == None


class TestsCLI:
    def test_home_cmd(self):
        self.setup_home()
        home_files_exist()

    def test_project_creation(self):
        new_project = tmp_config_path+'projects/tmp/hello.py'
        code = ['from xframe.interfaces import ProjectWorkerInterface\n',
                '\n',
                'class ProjectWorker(ProjectWorkerInterface):\n',
                '    def run(self):\n',
                "        print('Hello There!')\n",
                ]
        create_path_if_nonexistent(new_project)
        save(new_project,code)
        command = f"xframe tmp hello"
        p = subprocess.run(command,shell=True,capture_output=True)
        assert 'Hello There!' in str(p.stdout)

    def test_settings(self):
        ## Has to be executed after test_project_creation
        
        new_project = tmp_config_path+'projects/tmp/hello.py'
        settings_path_1 = tmp_config_path+'projects/tmp/settings/hello/opt1.yaml'
        settings_path_2 = tmp_config_path+'settings/projects/tmp/hello/opt2.yaml'
        code = ['from xframe.interfaces import ProjectWorkerInterface\n',
                'from xframe import settings\n',
                '\n',
                'class ProjectWorker(ProjectWorkerInterface):\n',
                '    def run(self):\n',
                "        opt = settings.project\n",
                "        print(f'Hello {opt.int}!')\n",
                ]
        settings = ['int:\n',
                    '  command: "int(42)"\n',
                    ]
        save(new_project,code)
        create_path_if_nonexistent(settings_path_1)
        save(settings_path_1,settings)
        create_path_if_nonexistent(settings_path_2)
        save(settings_path_2,settings)

        for i in range(1,3):
            cmd =  f"xframe tmp hello opt{i}"
            p = subprocess.run(cmd,shell=True,capture_output=True)
            assert 'Hello 42!' in str(p.stdout)

    def test_database(self):
        ## Has to be executed after test_project_creation
        
        new_project = tmp_config_path+'projects/tmp/hello.py'
        settings_path_1 = tmp_config_path+'projects/tmp/settings/hello/opt1.yaml'
        data_path = tmp_config_path+'data/tmp{val}.h5'
        code = ['from xframe.interfaces import ProjectWorkerInterface\n',
                'from xframe import database\n',
                '\n',
                'class ProjectWorker(ProjectWorkerInterface):\n',
                '    def run(self):\n',
                "        db = database.project\n",
                "        test_data={'answer':42}\n",
                "        path_modifiers={'val':0}\n",
                "        db.save('my_file',test_data,path_modifiers=path_modifiers)\n",
                "        loaded_file = db.load('my_file',path_modifiers=path_modifiers)\n",
                "        answer = loaded_file['answer']\n",
                "        print(f'Answer = {answer}')\n",
                ]
        settings = ['int:\n',
                    '  command: "int(42)"\n',
                    'IO:\n',
                    '  folders:\n',
                    f'    data: "{os.path.dirname(data_path)}/"\n',
                    '  files:\n',
                    '    my_file:\n',
                    f'     name: "{os.path.basename(data_path)}"\n',
                    '     folder: data\n',
                    ]
        save(new_project,code)
        create_path_if_nonexistent(settings_path_1)
        save(settings_path_1,settings)
        cmd =  f"xframe tmp hello opt1"
        p = subprocess.run(cmd,shell=True,capture_output=True)
        assert os.path.exists(data_path.format(val=0))==True
        assert 'Answer = 42' in str(p.stdout)

    def test_custom_database(self):
        ## Has to be executed after test_project_creation
        
        new_project = tmp_config_path+'projects/tmp/hello.py'
        settings_path_1 = tmp_config_path+'projects/tmp/settings/hello/opt1.yaml'
        data_path = tmp_config_path+'data/tmp.h5'
        db_path = tmp_config_path+'/projects/tmp/_database_.py'
        code = ['from xframe.interfaces import ProjectWorkerInterface\n',
                'from xframe import database\n',
                '\n',
                'class ProjectWorker(ProjectWorkerInterface):\n',
                '    def run(self):\n',
                "        db = database.project\n",
                "        test_data={'answer':42}\n",
                "        db.save('my_file',test_data)\n",
                "        loaded_file = db.load('my_file')\n",
                "        answer = loaded_file['answer']\n",
                "        yay = loaded_file['new_load']\n",
                "        nay = loaded_file['new_save']\n",
                "        print(f'Answer = {answer} {yay} {nay}')\n",
                ]
        db_code = [
            "from xframe.database.database import DefaultDB\n",
            "from xframe.interfaces import DatabaseInterface\n",
            "\n",
            "class ProjectDB(DefaultDB,DatabaseInterface):\n",
            "    def load_my_file(self,name,**kwargs):\n",
            "        data = self.load_direct(name,**kwargs)\n",
            "        data['new_load'] = 'yay'\n",
            "        return data\n",
            "\n",
            "    def save_my_file(self,name,data,**kwargs):\n",
            "        data['new_save'] = 'ney'\n",
            "        data = self.save_direct(name,data,**kwargs)\n",            
        ]
        settings = ['int:\n',
                    '  command: "int(42)"\n',
                    'IO:\n',
                    '  folders:\n',
                    f'    data: "{os.path.dirname(data_path)}/"\n',
                    '  files:\n',
                    '    my_file:\n',
                    f'     name: "{os.path.basename(data_path)}"\n',
                    '     folder: data\n',
                    ]
        save(new_project,code)
        create_path_if_nonexistent(settings_path_1)
        save(settings_path_1,settings)
        create_path_if_nonexistent(db_path)
        save(db_path,db_code)
        cmd =  f"xframe tmp hello opt1"
        p = subprocess.run(cmd,shell=True,capture_output=True)
        #assert os.path.exists(data_path)==True
        assert 'Answer = 42 yay ney' in str(p.stdout)
      
    def test_simple_multiprocessing(self):
        new_project = tmp_config_path+'projects/tmp/mp.py'
        code = [
            "from xframe.interfaces import ProjectWorkerInterface\n",
            "from xframe import Multiprocessing\n",
            "import numpy as np\n",
            "\n",
            "def multiply_matrix_with_vectors(vects,matrix,**kwargs):\n",
            "    vects= np.atleast_2d(vects)\n",
            "    new_vects = np.sum(matrix[None,:,:]*vects[:,None,:],axis=2)\n",
            "    return np.squeeze(new_vects)\n",
            "\n",
            "class ProjectWorker(ProjectWorkerInterface):\n",
            "    def run(self):\n",
            "        vectors = np.random.rand(200,10)\n",
            "        matrix = np.random.rand(10,10)\n",
            "\n",
            "        result = Multiprocessing.process_mp_request(multiply_matrix_with_vectors,input_arrays=[vectors],const_inputs = [matrix])\n",
            "\n",
            "        test_result = multiply_matrix_with_vectors(vectors,matrix)\n",
            "        if (result == test_result).all():\n",
            "            print('Test passed!')\n",            
        ]
        save(new_project,code)
        cmd =  f"xframe tmp mp"
        p = subprocess.run(cmd,shell=True,capture_output=True)
        assert 'Test passed!' in str(p.stdout)        
    def test_GPU(self):
        new_project = tmp_config_path+'projects/tmp/gpu.py'
        code = [
            "import numpy as np\n",
            "from xframe.interfaces import ProjectWorkerInterface\n",
            "from xframe import settings\n",
            "from xframe import Multiprocessing\n",
            "import xframe\n",
            "\n",
            "# Ensure a GPU worker is running\n",
            "settings.general.n_control_workers = 1\n",
            "xframe.controller.control_worker.restart_working()\n",
            "\n",
            "\n",
            "class ProjectWorker(ProjectWorkerInterface):\n",
            "    def run(self):\n",
            "\n",
            "        nq = 10 # vector length\n",
            "        nvec=5  # number of vectors\n",
            "        matrix = np.random.rand(nq,nq)\n",
            "        vects= np.random.rand(nq,nvec)\n",
            "        expected = matrix@vects # Expected result from module import symbol the gpu process\n",
            "\n",
            "        gpu_func = self.create_gpu_function(matrix,vects) # Create GPU Kernel function\n",
            "\n",
            "        result = gpu_func(vects) # Evaluate kernel on given vectors\n",
            "        if (result==expected).all(): # Check if gpu version gives same result as numpy computation.\n",
            "            print('Test passed!') \n",
            "\n",
            "    def create_gpu_function(self,matrix,vects):\n",
            "        nq,nvec=vects.shape\n",
            "        # Define Kernel\n",
            '        kernel_str = """\n',
            "            __kernel void\n",
            "            apply_matrix(__global double* out, \n",
            "            __global double* matrix, \n",
            "            __global double* vect, \n",
            "            long nq,long nvec)\n",
            "            {\n",
            "            long i = get_global_id(0); \n",
            "            long j = get_global_id(1);\n",
            "\n",
            "            // Compute application of i'th matrix row on j'th vector\n",
            "            // Store result in value\n",
            "            double value = 0;\n",
            "            for (int q = 0; q < nq; ++q)\n",
            "            {\n",
            "            double matqq = matrix[i*nq + q];\n",
            "            double veciq = vect[q*nvec + j];\n",
            "            value += matqq * veciq;\n",
            "            }\n",
            "\n",
            "            // Write the result vector to device memory\n",
            "            out[i * nvec + j] = value;\n",
            "            }\n",
            '            """\n',
            "        # Define types and input constant arguments\n",
            "        kernel_dict_forward={\n",
            "            'kernel': kernel_str,\n",
            "            'name': 'gpu_func',\n",
            "            'functions': ({\n",
            "                'name': 'apply_matrix',\n",
            "                'dtypes' : (float,float,float,np.int64,np.int64),\n",
            "                'shapes' : ((nq,nvec),matrix.shape,(nq,nvec),None,None,None),\n",
            "                'arg_roles' : ('output','const_input','input','const_input','const_input'),\n",
            "                'const_inputs' : (None,matrix,None,np.int64(nq),np.int64(nvec)),\n",
            "                'global_range' : (nq,nvec),\n",
            "                'local_range' : None\n",
            "                },)\n",
            "            }\n",
            "        # Create cl Process\n",
            "        cl_process =  Multiprocessing.openCL_plugin.ClProcess(kernel_dict_forward)\n",
            "        # Register opencl function and get gpu client function\n",
            "        gpu_func = Multiprocessing.comm_module.add_gpu_process(cl_process)\n",
            "        return gpu_func\n",            
        ]
        save(new_project,code)
        cmd =  f"xframe tmp gpu"
        p = subprocess.run(cmd,shell=True,capture_output=True)
        assert 'Test passed!' in str(p.stdout)
        pass
    def test_GPU_in_multiprocessing(self):
        new_project = tmp_config_path+'projects/tmp/gpu.py'
        n_processes = 10
        code = [
            "import numpy as np\n",
            "from xframe.interfaces import ProjectWorkerInterface\n",
            "from xframe import settings\n",
            "from xframe import Multiprocessing\n",
            "import xframe\n",
            "\n",
            "# Ensure a GPU worker is running\n",
            "settings.general.n_control_workers = 1\n",
            "xframe.controller.control_worker.restart_working()\n",
            "\n",
            "\n",
            "class ProjectWorker(ProjectWorkerInterface):\n",
            "    def run(self):\n",
            "        nq = 10 # vector length\n",
            "        nvec=5  # number of vectors\n",
            "        matrix = np.random.rand(nq,nq)\n",
            "        n_processes = 10\n",
            "        def run_parallel(**kwargs):\n",
            "            vects= np.random.rand(nq,nvec)\n",
            "            gpu_func = self.create_gpu_function(matrix,vects) # Create GPU Kernel function\n",
            "            expected = matrix@vects # Expected result from the gpu process\n",
            "            result = gpu_func(vects) # Evaluate kernel on given vectors\n",
            "            if (result==expected).all(): # Check if gpu version gives same result as numpy computation.\n",
            "                print(f'Process {Multiprocessing.get_process_name()}: Test passed!')\n",
            f"        Multiprocessing.process_mp_request(run_parallel,n_processes = {n_processes})\n",
            "\n",
            "    def create_gpu_function(self,matrix,vects):\n",
            "        nq,nvec=vects.shape\n",
            "        # Define Kernel\n",
            '        kernel_str = """\n',
            "            __kernel void\n",
            "            apply_matrix(__global double* out, \n",
            "            __global double* matrix, \n",
            "            __global double* vect, \n",
            "            long nq,long nvec)\n",
            "            {\n",
            "            long i = get_global_id(0); \n",
            "            long j = get_global_id(1);\n",
            "\n",
            "            // Compute application of i'th matrix row on j'th vector\n",
            "            // Store result in value\n",
            "            double value = 0;\n",
            "            for (int q = 0; q < nq; ++q)\n",
            "            {\n",
            "            double matqq = matrix[i*nq + q];\n",
            "            double veciq = vect[q*nvec + j];\n",
            "            value += matqq * veciq;\n",
            "            }\n",
            "\n",
            "            // Write the result vector to device memory\n",
            "            out[i * nvec + j] = value;\n",
            "            }\n",
            '            """\n',
            "        # Define types and input constant arguments\n",
            "        kernel_dict_forward={\n",
            "            'kernel': kernel_str,\n",
            "            'name': 'gpu_func',\n",
            "            'functions': ({\n",
            "                'name': 'apply_matrix',\n",
            "                'dtypes' : (float,float,float,np.int64,np.int64),\n",
            "                'shapes' : ((nq,nvec),matrix.shape,(nq,nvec),None,None,None),\n",
            "                'arg_roles' : ('output','const_input','input','const_input','const_input'),\n",
            "                'const_inputs' : (None,matrix,None,np.int64(nq),np.int64(nvec)),\n",
            "                'global_range' : (nq,nvec),\n",
            "                'local_range' : None\n",
            "                },)\n",
            "            }\n",
            "        # Create cl Process\n",
            "        cl_process =  Multiprocessing.openCL_plugin.ClProcess(kernel_dict_forward)\n",
            "        # Register opencl function and get gpu client function\n",
            "        gpu_func = Multiprocessing.comm_module.add_gpu_process(cl_process)\n",
            "        return gpu_func\n",            
        ]
        save(new_project,code)
        cmd =  f"xframe tmp gpu"
        p = subprocess.run(cmd,shell=True,capture_output=True)
        for process in np.arange(1,n_processes+1):
            assert f'Process {process}: Test passed!' in str(p.stdout)

    
    def setup_home(self):
        print(f'new home at : {tmp_config_path}')
        print(f'given path exists: {os.path.exists(_path)}')
        command = f"xframe --setup_home {tmp_config_path}"
        p = subprocess.Popen(command,shell=True,stdin=subprocess.PIPE)
        p.communicate('y'.encode())[0]

class TestsScripting:
    def test_home_cmd(self):
        xframe.change_home(tmp_config_path)
        home_files_exist()
    def test_project_creation(self,capsys):
        global tmp_config_path
        new_project = tmp_config_path+'projects/tmp/hello.py'
        code = ['from xframe.interfaces import ProjectWorkerInterface\n',
                '\n',
                'class ProjectWorker(ProjectWorkerInterface):\n',
                '    def run(self):\n',
                "        print('Hello There!')\n",
                ]
        create_path_if_nonexistent(new_project)
        save(new_project,code)
        xframe.reload_home()
        
        xframe.select_project('tmp','hello')        
        xframe.run(update_worker=True)
        captured = capsys.readouterr()
        assert 'Hello There!' in captured.out
    def test_settings(self,capsys):
        ## Has to be executed after test_project_creation
        
        new_project = tmp_config_path+'projects/tmp/hello.py'
        settings_path_1 = tmp_config_path+'projects/tmp/settings/hello/opt1.yaml'
        settings_path_2 = tmp_config_path+'settings/projects/tmp/hello/opt2.yaml'
        code = ['from xframe.interfaces import ProjectWorkerInterface\n',
                'from xframe import settings\n',
                '\n',
                'class ProjectWorker(ProjectWorkerInterface):\n',
                '    def run(self):\n',
                "        opt = settings.project\n",
                "        print(f'Hello {opt.int}!')\n",
                ]
        settings = ['int:\n',
                    '  command: "int(42)"\n',
                    ]
        save(new_project,code)
        create_path_if_nonexistent(settings_path_1)
        save(settings_path_1,settings)
        create_path_if_nonexistent(settings_path_2)
        save(settings_path_2,settings)
        xframe.reload_home()
        for i in range(1,3):
            xframe.select_project('tmp','hello',f'opt{i}')
            xframe.run()
            captured = capsys.readouterr()
            assert 'Hello 42!' in str(captured.out)

    def test_database(self,capsys):
        ## Has to be executed after test_project_creation
        
        new_project = tmp_config_path+'projects/tmp/hello.py'
        settings_path_1 = tmp_config_path+'projects/tmp/settings/hello/opt1.yaml'
        data_path = tmp_config_path+'data/tmp{val}.h5'
        code = ['from xframe.interfaces import ProjectWorkerInterface\n',
                'from xframe import database\n',
                '\n',
                'class ProjectWorker(ProjectWorkerInterface):\n',
                '    def run(self):\n',
                "        db = database.project\n",
                "        test_data={'answer':42}\n",
                "        path_modifiers={'val':0}\n",
                "        db.save('my_file',test_data,path_modifiers=path_modifiers)\n",
                "        loaded_file = db.load('my_file',path_modifiers=path_modifiers)\n",
                "        answer = loaded_file['answer']\n",
                "        print(f'Answer = {answer}')\n",
                ]
        settings = ['int:\n',
                    '  command: "int(42)"\n',
                    'IO:\n',
                    '  folders:\n',
                    f'    data: "{os.path.dirname(data_path)}/"\n',
                    '  files:\n',
                    '    my_file:\n',
                    f'     name: "{os.path.basename(data_path)}"\n',
                    '     folder: data\n',
                    ]
        save(new_project,code)
        create_path_if_nonexistent(settings_path_1)
        save(settings_path_1,settings)
        xframe.reload_home()
        xframe.select_project('tmp','hello',f'opt1')
        xframe.run()
        captured = capsys.readouterr()
        assert os.path.exists(data_path.format(val=0))==True
        assert 'Answer = 42' in str(captured.out)

    def test_custom_database(self,capsys):
        ## Has to be executed after test_project_creation
        
        new_project = tmp_config_path+'projects/tmp/hello.py'
        settings_path_1 = tmp_config_path+'projects/tmp/settings/hello/opt1.yaml'
        data_path = tmp_config_path+'data/tmp.h5'
        db_path = tmp_config_path+'/projects/tmp/_database_.py'
        code = ['from xframe.interfaces import ProjectWorkerInterface\n',
                'from xframe import database\n',
                '\n',
                'class ProjectWorker(ProjectWorkerInterface):\n',
                '    def run(self):\n',
                "        db = database.project\n",
                "        test_data={'answer':42}\n",
                "        db.save('my_file',test_data)\n",
                "        loaded_file = db.load('my_file')\n",
                "        answer = loaded_file['answer']\n",
                "        yay = loaded_file['new_load']\n",
                "        nay = loaded_file['new_save']\n",
                "        print(f'Answer = {answer} {yay} {nay}')\n",
                ]
        db_code = [
            "from xframe.database.database import DefaultDB\n",
            "from xframe.interfaces import DatabaseInterface\n",
            "\n",
            "class ProjectDB(DefaultDB,DatabaseInterface):\n",
            "    def load_my_file(self,name,**kwargs):\n",
            "        data = self.load_direct(name,**kwargs)\n",
            "        data['new_load'] = 'yay'\n",
            "        return data\n",
            "\n",
            "    def save_my_file(self,name,data,**kwargs):\n",
            "        data['new_save'] = 'ney'\n",
            "        data = self.save_direct(name,data,**kwargs)\n",            
        ]
        settings = ['int:\n',
                    '  command: "int(42)"\n',
                    'IO:\n',
                    '  folders:\n',
                    f'    data: "{os.path.dirname(data_path)}/"\n',
                    '  files:\n',
                    '    my_file:\n',
                    f'     name: "{os.path.basename(data_path)}"\n',
                    '     folder: data\n',
                    ]
        save(new_project,code)
        create_path_if_nonexistent(settings_path_1)
        save(settings_path_1,settings)
        create_path_if_nonexistent(db_path)
        save(db_path,db_code)
        xframe.reload_home()
        xframe.select_project('tmp','hello','opt1')
        xframe.run()
        captured = capsys.readouterr()
        assert 'Answer = 42 yay ney' in str(captured.out)
      
    def test_simple_multiprocessing(self,capsys):
        new_project = tmp_config_path+'projects/tmp/mp.py'
        code = [
            "from xframe.interfaces import ProjectWorkerInterface\n",
            "from xframe import Multiprocessing\n",
            "import numpy as np\n",
            "\n",
            "def multiply_matrix_with_vectors(vects,matrix,**kwargs):\n",
            "    vects= np.atleast_2d(vects)\n",
            "    new_vects = np.sum(matrix[None,:,:]*vects[:,None,:],axis=2)\n",
            "    return np.squeeze(new_vects)\n",
            "\n",
            "class ProjectWorker(ProjectWorkerInterface):\n",
            "    def run(self):\n",
            "        vectors = np.random.rand(200,10)\n",
            "        matrix = np.random.rand(10,10)\n",
            "\n",
            "        result = Multiprocessing.process_mp_request(multiply_matrix_with_vectors,input_arrays=[vectors],const_inputs = [matrix])\n",
            "\n",
            "        test_result = multiply_matrix_with_vectors(vectors,matrix)\n",
            "        if (result == test_result).all():\n",
            "            print('Test passed!')\n",            
        ]
        save(new_project,code)
        xframe.reload_home()
        xframe.select_project('tmp','mp')
        xframe.run()
        captured = capsys.readouterr()
        assert 'Test passed!' in str(captured.out)        
    def test_GPU(self,capsys):
        new_project = tmp_config_path+'projects/tmp/gpu.py'
        code = [
            "import numpy as np\n",
            "from xframe.interfaces import ProjectWorkerInterface\n",
            "from xframe import settings\n",
            "from xframe import Multiprocessing\n",
            "import xframe\n",
            "\n",
            "# Ensure a GPU worker is running\n",
            "settings.general.n_control_workers = 1\n",
            "xframe.controller.control_worker.restart_working()\n",
            "\n",
            "\n",
            "class ProjectWorker(ProjectWorkerInterface):\n",
            "    def run(self):\n",
            "\n",
            "        nq = 10 # vector length\n",
            "        nvec=5  # number of vectors\n",
            "        matrix = np.random.rand(nq,nq)\n",
            "        vects= np.random.rand(nq,nvec)\n",
            "        expected = matrix@vects # Expected result from module import symbol the gpu process\n",
            "\n",
            "        gpu_func = self.create_gpu_function(matrix,vects) # Create GPU Kernel function\n",
            "\n",
            "        result = gpu_func(vects) # Evaluate kernel on given vectors\n",
            "        if (result==expected).all(): # Check if gpu version gives same result as numpy computation.\n",
            "            print('Test passed!') \n",
            "\n",
            "    def create_gpu_function(self,matrix,vects):\n",
            "        nq,nvec=vects.shape\n",
            "        # Define Kernel\n",
            '        kernel_str = """\n',
            "            __kernel void\n",
            "            apply_matrix(__global double* out, \n",
            "            __global double* matrix, \n",
            "            __global double* vect, \n",
            "            long nq,long nvec)\n",
            "            {\n",
            "            long i = get_global_id(0); \n",
            "            long j = get_global_id(1);\n",
            "\n",
            "            // Compute application of i'th matrix row on j'th vector\n",
            "            // Store result in value\n",
            "            double value = 0;\n",
            "            for (int q = 0; q < nq; ++q)\n",
            "            {\n",
            "            double matqq = matrix[i*nq + q];\n",
            "            double veciq = vect[q*nvec + j];\n",
            "            value += matqq * veciq;\n",
            "            }\n",
            "\n",
            "            // Write the result vector to device memory\n",
            "            out[i * nvec + j] = value;\n",
            "            }\n",
            '            """\n',
            "        # Define types and input constant arguments\n",
            "        kernel_dict_forward={\n",
            "            'kernel': kernel_str,\n",
            "            'name': 'gpu_func',\n",
            "            'functions': ({\n",
            "                'name': 'apply_matrix',\n",
            "                'dtypes' : (float,float,float,np.int64,np.int64),\n",
            "                'shapes' : ((nq,nvec),matrix.shape,(nq,nvec),None,None,None),\n",
            "                'arg_roles' : ('output','const_input','input','const_input','const_input'),\n",
            "                'const_inputs' : (None,matrix,None,np.int64(nq),np.int64(nvec)),\n",
            "                'global_range' : (nq,nvec),\n",
            "                'local_range' : None\n",
            "                },)\n",
            "            }\n",
            "        # Create cl Process\n",
            "        cl_process =  Multiprocessing.openCL_plugin.ClProcess(kernel_dict_forward)\n",
            "        # Register opencl function and get gpu client function\n",
            "        gpu_func = Multiprocessing.comm_module.add_gpu_process(cl_process)\n",
            "        return gpu_func\n",            
        ]
        save(new_project,code)
        xframe.reload_home()
        xframe.select_project('tmp','gpu')
        xframe.run()
        captured = capsys.readouterr()
        assert 'Test passed!' in str(captured.out)
    def test_GPU_in_multiprocessing(self,capsys):
        new_project = tmp_config_path+'projects/tmp/gpu.py'
        n_processes = 5
        code = [
            "import numpy as np\n",
            "from xframe.interfaces import ProjectWorkerInterface\n",
            "from xframe import settings\n",
            "from xframe import Multiprocessing\n",
            "import xframe\n",
            "\n",
            "# Ensure a GPU worker is running\n",
            "settings.general.n_control_workers = 1\n",
            "xframe.controller.control_worker.restart_working()\n",
            "\n",
            "\n",
            "class ProjectWorker(ProjectWorkerInterface):\n",
            "    def run(self):\n",
            "        nq = 10 # vector length\n",
            "        nvec=5  # number of vectors\n",
            "        matrix = np.random.rand(nq,nq)\n",
            "        n_processes = 10\n",
            "        def run_parallel(**kwargs):\n",
            "            vects= np.random.rand(nq,nvec)\n",
            "            gpu_func = self.create_gpu_function(matrix,vects) # Create GPU Kernel function\n",
            "            expected = matrix@vects # Expected result from the gpu process\n",
            "            result = gpu_func(vects) # Evaluate kernel on given vectors\n",
            "            if (result==expected).all(): # Check if gpu version gives same result as numpy computation.\n",            
            "                return True\n",
            "            else:\n",
            "                return False\n",
            f"        _out = Multiprocessing.process_mp_request(run_parallel,n_processes = {n_processes})\n",
            "        if all(_out):\n",
            "             print('Test passed!')\n",
            "\n",
            "    def create_gpu_function(self,matrix,vects):\n",
            "        nq,nvec=vects.shape\n",
            "        # Define Kernel\n",
            '        kernel_str = """\n',
            "            __kernel void\n",
            "            apply_matrix(__global double* out, \n",
            "            __global double* matrix, \n",
            "            __global double* vect, \n",
            "            long nq,long nvec)\n",
            "            {\n",
            "            long i = get_global_id(0); \n",
            "            long j = get_global_id(1);\n",
            "\n",
            "            // Compute application of i'th matrix row on j'th vector\n",
            "            // Store result in value\n",
            "            double value = 0;\n",
            "            for (int q = 0; q < nq; ++q)\n",
            "            {\n",
            "            double matqq = matrix[i*nq + q];\n",
            "            double veciq = vect[q*nvec + j];\n",
            "            value += matqq * veciq;\n",
            "            }\n",
            "\n",
            "            // Write the result vector to device memory\n",
            "            out[i * nvec + j] = value;\n",
            "            }\n",
            '            """\n',
            "        # Define types and input constant arguments\n",
            "        kernel_dict_forward={\n",
            "            'kernel': kernel_str,\n",
            "            'name': 'gpu_func',\n",
            "            'functions': ({\n",
            "                'name': 'apply_matrix',\n",
            "                'dtypes' : (float,float,float,np.int64,np.int64),\n",
            "                'shapes' : ((nq,nvec),matrix.shape,(nq,nvec),None,None,None),\n",
            "                'arg_roles' : ('output','const_input','input','const_input','const_input'),\n",
            "                'const_inputs' : (None,matrix,None,np.int64(nq),np.int64(nvec)),\n",
            "                'global_range' : (nq,nvec),\n",
            "                'local_range' : None\n",
            "                },)\n",
            "            }\n",
            "        # Create cl Process\n",
            "        cl_process =  Multiprocessing.openCL_plugin.ClProcess(kernel_dict_forward)\n",
            "        # Register opencl function and get gpu client function\n",
            "        gpu_func = Multiprocessing.comm_module.add_gpu_process(cl_process)\n",
            "        return gpu_func\n",            
        ]
        save(new_project,code)
        xframe.reload_home()
        xframe.select_project('tmp','gpu')
        #with capsys.disabled():
        xframe.run()
        captured = capsys.readouterr()
        assert f'Test passed!' in str(captured.out)


